

// Weight
#include <XSStat/Weight.h>
// FitStatMethod
#include <XSStat/FitStatMethod.h>

#include <XSModel/Model/Model.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/DataUtility.h>
#include <XSModel/Data/Detector/Response.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/Parameter/ModParam.h>
#include <XSModel/Parameter/ResponseParam.h>
#include <XSStat/ChiSquare.h>
#include <XSStat/Cstat.h>
#include <XSStat/CstatVariants.h>
#include <XSUtil/Utils/XSutility.h>
#include <XSstreams.h>
#include <XSContainer.h>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <set>



// Class FitStatMethod::InvalidSpectrumNumber 

FitStatMethod::InvalidSpectrumNumber::InvalidSpectrumNumber (size_t num)
  : RedAlert(" initializing spectra for fit:\n ")  
{
  std::ostringstream message;
  message << " Statistic tried to access spectrum number " << num << " not loaded ";
  reportAndExit(message.str(),-5);
}


// Class FitStatMethod::InvalidParameterError 

FitStatMethod::InvalidParameterError::InvalidParameterError ()
  : RedAlert(" retrieving model derivative array:\n")              
{
  std::ostringstream message;
  message << " Statistic calculation attempted to retrieve parameter derivative not allocated ";
  reportAndExit(message.str(),-5);
}


// Class FitStatMethod::FitNotDefined 

FitStatMethod::FitNotDefined::FitNotDefined (const string& diag)
  : YellowAlert (" no model defined or no data read ")
{
  tcerr << diag << '\n';
}


// Class FitStatMethod::NoSuchSpectrum 

FitStatMethod::NoSuchSpectrum::NoSuchSpectrum (const string& diag)
  : YellowAlert(" spectrum not loaded ")
{
}


// Class FitStatMethod 


FitStatMethod::FitStatMethod (const std::string& name)
  : StatMethod(name),
    m_canUseChainRule(true)
{
}


FitStatMethod::~FitStatMethod()
{
}



bool FitStatMethod::checkWeight (const Weight* weightMethod) const
{
  //  static const string STANDARD ("standard");
  // Default checkWeight only accepts standard weighting.
  //  const string weightName = weightMethod->name();
  //  return STANDARD.find(weightName) == 0;
  // Allow all Weight methods
  return true;
}

void FitStatMethod::perform(const SpectralData* sp)
{
  doPerform(sp);
}

void FitStatMethod::doPerform(const SpectralData* sp)
{
}

void FitStatMethod::initialize (const SpectralData* sp)
{

  doInitialize(sp);
}

void FitStatMethod::doInitialize (const SpectralData* sp)
{

  // clear all and start again.
  deallocate();
  // allocate memory for derivative arrays. If derivatives are not
  // required we don't need to do anything.

  // create proto workspace

  size_t N = sp->indirectNotice().size();
  m_protoWorkspace.resize(N);
  m_protoWorkspace = 0.0;

}

void FitStatMethod::reset (const SpectralData* sp)
{

  doReset(sp);
}

void FitStatMethod::reportWarnings (const SpectralData* sp)
{

  doReportWarnings(sp);
}

void FitStatMethod::deallocate ()
{

        doDeallocate();
}

void FitStatMethod::doDeallocate ()
{
        m_protoWorkspace.resize(0); 
}

size_t FitStatMethod::nBinsUsed (const SpectralData* sp) const
{
    return sp->indirectNotice().size();
}

Real FitStatMethod::sumDerivs (const RealArray& dFolded, const int spectrumNumber, 
			       const int parameterIndex) const
{
   return .0;
}

Real FitStatMethod::statValueForSingleBin (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const
{

  return 0;
}

Real FitStatMethod::deltaSigmaForSingleBin (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const
{

  return 0;
}

Real FitStatMethod::sumSecondDerivs (const std::map<size_t,RealArray>& dFolded, int pi, int pj)
{
   return .0;
}

void FitStatMethod::getWeightDerivs(const SpectralData* spec)
{
}

FitStatMethod* FitStatMethod::createFitStatMethodObject(const string& name, const int& number, const string& modifier)
{
  // this method creates and returns the pointer to the appropriate new object.

  static string chi("chi");
  static string cstat("cstat");
  static string lstat("lstat");
  static string pgstat("pgstat");
  static string pstat("pstat");
  static string whittle("whittle");

  FitStatMethod* stat = 0;
  if ( name == chi.substr(0,name.length()) ) { // chi
    // if a modifier string is included check that it is a valid weighting name
    if ( modifier.length() > 0 ) {
      string weightMethod = modifier;
      if ( Weight::isValidWeightName(weightMethod) == false ) {
	string errStr = modifier + " is not a valid weighting for chi-squared\n";
	throw YellowAlert(errStr);
      }
    }
    stat = new ChiSquare;
  } else if ( name == cstat.substr(0,name.length()) ) { // cstat
    stat = new Cstat<StdCstat>;
  } else if ( name == lstat.substr(0,name.length()) ) { // lstat
    stat = new Cstat<LorStat>;
  } else if ( name == pgstat.substr(0,name.length()) ) { // pgstat
    stat = new Cstat<PGstat>;
  } else if ( name == pstat.substr(0,name.length()) ) { // pstat
    stat = new Cstat<Pstat>;
  } else if ( name == whittle.substr(0,name.length() ) ) { // whittle
    stat = new Cstat<WhittleStat>;
  } else {
    return stat;
  }

  if ( number != -999 ) stat->number(number);
  if ( modifier != "" ) stat->modifier(modifier);

  return stat;
}

bool FitStatMethod::isValidStatName(const string& name, const int& numeral, const string& modifier)
{
  bool valid(false);

  // try to create a FitStatMethod object
  try {
    FitStatMethod* stat = createFitStatMethodObject(name, numeral, modifier);
    // if the created stat exists then input is valid
    if (stat) valid = true;
    // tidy up
    delete stat;
  } catch(...) {
    valid = false;
  }

  return valid;
}

string FitStatMethod::statNames ()
{
  return "chi | cstat | lstat | pgstat | pstat | whittle";
}




// Additional Declarations
