//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%

// Model
#include <XSModel/Model/Model.h>
// SumComponent
#include <XSModel/Model/Component/SumComponent.h>
// MultiBinningResponse
#include <XSModel/Data/Detector/MultiBinningResponse.h>
#include <XSContainer.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/Parameter/ResponseParam.h>
#include <GlobalContainer/ResponseContainer.h>
#include <Data/BackCorr/Background.h>
#include <XSUtil/Numerics/LinearInterp.h>
#include <XSstreams.h>
#include <cstring>
#include <memory>

#include "rmf.h"


// Class MultiBinningResponse::positionInfo 

// Class MultiBinningResponse 

MultiBinningResponse::MultiBinningResponse()
      : Response(),
	m_effectiveArea(),
        m_responses(),
        m_responseBinFactors()
{
}

MultiBinningResponse::MultiBinningResponse(const MultiBinningResponse &right)
      : Response(right),
        m_effectiveArea(right.m_effectiveArea),
        m_responses(right.m_responses),
        m_responseBinFactors(right.m_responseBinFactors)
        
{
}


MultiBinningResponse::~MultiBinningResponse()
{
  // NB need to add destructors for each individual member of m_responses.
  m_effectiveArea.resize(0);
  m_responses.resize(0);
  m_responseBinFactors.resize(0);
}


Model& MultiBinningResponse::operator * (Model& model) const
{
   int iData(spectrumNumber());

   const RealArray& modFlux = model.modelFlux(iData);
   RealArray foldedMod(0.,numChannels());
   if ( modFlux.size())
   {
           const RealArray& modFluxErr = model.modelFluxError(iData);
           RealArray foldedModErr;
           if (modFluxErr.size()) foldedModErr.resize(numChannels());
           convolve(modFlux,modFluxErr,foldedMod,foldedModErr);

// Can't have multiple threads trying to insert arrays into the
// same Model object.  That would cause a race condition.
#pragma omp critical
          {
           model.foldedModel(iData,foldedMod);
           if (modFluxErr.size()) model.foldedModelError(iData,foldedModErr);
          }
   }    
   else
   {
#pragma omp critical
          {
           model.foldedModel(iData,foldedMod);
           model.foldedModelError(iData,RealArray());
          }           
   }           
   return model;
}

void MultiBinningResponse::setNumberResponses ( const string fileName )
{
  // this initializes the responses and responseBinFactors arrays
  // by checking for the number of MATRIX extensions in fileName.
  // if fileName has a {n} suffix then set the number of response to one.

  size_t nRmf(1);
  if ( fileName.find_first_of('{') == string::npos ) {
    IntegerVector extNums = RMFmatrixExtensions(fileName);
    nRmf = extNums.size();
  }

  m_responses.resize(nRmf);
  m_responseBinFactors.resize(nRmf);
  for (size_t i=0; i<nRmf; i++) m_responseBinFactors[i] = 1;
}

void MultiBinningResponse::setData (size_t specNum, size_t groupNumber)
{
  // this should not be possible
  if ( m_responses.size() == 0 ) return;

  for (size_t i=0; i<m_responses.size(); i++) {
    m_responses[i].setData(specNum, groupNumber);
  }

  // the response m_energies array should be set to the largest array of
  // the constituent responses
  size_t maxSizeIndex(0);
  size_t maxSizeArray(m_responses[0].numEnergies()+1);
  for (size_t i=1; i<m_responses.size(); i++) {
    if ( m_responses[i].numEnergies()+1 > maxSizeArray ) {
      maxSizeIndex = i;
      maxSizeArray = m_responses[i].numEnergies()+1;
    }
  }

  energies().resize(maxSizeArray);
  for (size_t i=0; i<maxSizeArray; i++) {
    energies(i,m_responses[maxSizeIndex].energies(i));
  }

  // now set the response bin factors relative to the longest array
  for (size_t i=0; i<m_responses.size(); i++) {
    m_responseBinFactors[i] = (maxSizeArray-1) / m_responses[i].numEnergies();
  }

  setSharedDescription(specNum, groupNumber);
}

bool MultiBinningResponse::readAuxResponse (int rowNum)
{
  bool isSuccess(true);
  m_effectiveArea.resize(numEnergies());
  m_effectiveArea = 0.0;
  for (size_t i=0; i<m_responses.size(); i++) {
    if ( !m_responses[i].readAuxResponse(rowNum) ) {
      isSuccess = false;
    } else {
      const RealArray& singleEff = m_responses[i].efficiency();
      for (size_t ie=0; ie<m_effectiveArea.size(); ie++) {
	m_effectiveArea[ie] += singleEff[ie/m_responseBinFactors[i]];
      }
    }
  }
  return isSuccess;
}

bool MultiBinningResponse::fileFormat (const string& fileName, XspecDataIO::DataType type)
{
  string testFile(fileName);
  return isGoodResponseMatrixFile(testFile);
}

bool MultiBinningResponse::isGoodFile(string& fileName)
{
  // check whether this is a good response matrix. if a [n] suffix is
  // present then it is converted to {n}
  return isGoodResponseMatrixFile(fileName);
}


void MultiBinningResponse::setSharedDescription (size_t specNum, size_t groupNum)
{
  size_t nChan=0;
  size_t nEner=0;
  for (size_t i=0; i<m_responses.size(); i++) {
    if ( m_responses[i].numChannels() > nChan ) nChan = m_responses[i].numChannels();
    if ( m_responses[i].numEnergies() > nEner ) nEner = m_responses[i].numEnergies();
  }
  numChannels(nChan);
  numEnergies(nEner);

  dataGroup(groupNum);
  spectrumNumber(specNum);

  // sum up the total effective area which will have been set in the
  // setSharedDescription call for individual SingleResponse.

  m_effectiveArea.resize(numEnergies());
  m_effectiveArea = 0.0;
  for (size_t i=0; i<m_responses.size(); i++) {
    const RealArray& singleEff = m_responses[i].efficiency();
    size_t binFactor = m_responseBinFactors[i];
    for (size_t ie=0; ie<m_effectiveArea.size(); ie++) {
      m_effectiveArea[ie] += singleEff[ie/binFactor];
    }
  }

}

void MultiBinningResponse::RMFremove ()
{
  for (size_t i=0; i<m_responses.size(); i++) m_responses[i].RMFremove();
}

void MultiBinningResponse::prepareForFit ()
{
  for (size_t i=0; i<m_responses.size(); i++) m_responses[i].prepareForFit();
}

const RealArray& MultiBinningResponse::eboundsMin () const
{
  return m_responses[0].eboundsMin();
}

const RealArray& MultiBinningResponse::eboundsMax () const
{
  return m_responses[0].eboundsMax();
}

void MultiBinningResponse::convolve (const RealArray& flux, const RealArray& fluxErr, RealArray& foldFlux, RealArray& foldFluxErr) const
{
   foldFlux = 0.0;
   if ( fluxErr.size() != 0 ) foldFluxErr = 0.0;

   // loop over the individual responses binning the flux (and fluxErr) arrays
   // as necessary

   for (size_t iresp=0; iresp<m_responses.size(); iresp++) {

     int binF = m_responseBinFactors[iresp];
     RealArray binnedFlux(0.0,flux.size()/binF);
     RealArray binnedFluxErr(0.0,fluxErr.size()/binF);

     if ( binF != 1 ) {

       size_t ibin(0);
       for (size_t i=0; i<flux.size(); i++) {
	 binnedFlux[ibin] += flux[i];
	 if ( (int)i % binF == binF - 1) ibin++;
       }

       if ( fluxErr.size() != 0 ) {
	 size_t ibin(0);
	 for (size_t i=0; i<flux.size(); i++) {
	   binnedFluxErr[ibin] += fluxErr[i]*fluxErr[i];
	   if ( (int)i % binF == binF - 1) ibin++;
	 }
	 binnedFluxErr = sqrt(binnedFluxErr);
       }

     } else {
       binnedFlux = flux;
       if ( fluxErr.size() != 0 ) binnedFluxErr = fluxErr;
     }

     // run the convolve for the current response

     RealArray tempFoldFlux, tempFoldFluxErr;
     m_responses[iresp].convolve(binnedFlux, binnedFluxErr, tempFoldFlux, tempFoldFluxErr);

     // and sum into the output

     foldFlux += tempFoldFlux;
     if ( fluxErr.size() != 0 ) foldFluxErr += tempFoldFluxErr * tempFoldFluxErr;

   }

   // if propagating errors then the accumulated folded flux error array is the
   // square of the required output.
   if ( fluxErr.size() != 0 ) foldFluxErr = sqrt(foldFluxErr);

}

RealArray MultiBinningResponse::sensitivity (const SpectralData* data)
{
  RealArray sens(0.,numEnergies());
  for (size_t iresp=0; iresp<m_responses.size(); iresp++) {
    sens += m_responses[iresp].sensitivity(data);
  }
  return sens;
}

void MultiBinningResponse::operator * (SumComponent& source) const
{
   int iData(spectrumNumber());
  // one of those rare uses for const_cast which seem to be acceptable.
  // source is cast to const so it can call the const versions of
  // photonArray(), etc ... the non-const versions are protected.
   const SumComponent& csource = const_cast<SumComponent&>(source);      
   const RealArray& photonFlux = csource.photonArray(iData);
   source.foldedPhotonFlux(iData).resize(numChannels(),0.);
   if ( csource.photonErrArray(iData).size())  
   {
           source.foldedPhotonFluxErr(iData).resize(numChannels(),0.);
   }
   convolve(photonFlux,csource.photonErrArray(iData),source.foldedPhotonFlux(iData),
                             source.foldedPhotonFluxErr(iData));
}

void MultiBinningResponse::shiftEffectiveArea (const RealArray& newEnergies)
{
  for (size_t i=0; i<m_responses.size(); i++) {
    m_responses[i].shiftEffectiveArea(newEnergies);
  }

  const Real FUZZY = 1.0e-6;
  // Lazy initialization of m_savedEffectiveArea array. If gain
  // command is never called, this need not be done.
  if (!m_savedEffectiveArea.size())
  {
     m_savedEffectiveArea.resize(m_effectiveArea.size());
     m_savedEffectiveArea = m_effectiveArea;
  }
  size_t inStart, outStart;
  if (!Numerics::Rebin::findFirstBins(energies(), newEnergies, FUZZY,
                inStart, outStart))
  {
     string msg("***No new energy bins overlap the old, cannot apply gain.");
     throw GainError(msg);
  }
  IntegerVector startBin, endBin;
  RealArray startWeight, endWeight;
  Numerics::Rebin::initializeBins(energies(), newEnergies, FUZZY, inStart,
                outStart, startBin, endBin, startWeight, endWeight);
  // this function modifies m_effectiveArea
  Numerics::Rebin::gainRebin(m_savedEffectiveArea, startBin, endBin, 
                startWeight, endWeight, m_effectiveArea);
}

void MultiBinningResponse::restoreEffectiveArea ()
{

  for (size_t i=0; i<m_responses.size(); i++) {
    m_responses[i].restoreEffectiveArea();
  }

  size_t sz = m_savedEffectiveArea.size();
  if (sz)
  {
     if (m_effectiveArea.size() == sz)
     {
        m_effectiveArea = m_savedEffectiveArea;
     }
     else
     {
        throw RedAlert("Size mismatch between old and new effectiveArea array.");
     }
  }
}

void MultiBinningResponse::calcEffAreaPerChan (RealArray& effArea)
{
   // Can't assume prepareForFit was called.  There may be no model
   // associated with this response.
   prepareForFit(); 
   const RealArray& ebins = energies();
   const size_t nBins = ebins.size()-1;
   RealArray dummyError;
   effArea.resize(numChannels(), .0);
   // flatFlux corresponds to a constant rate/keV.  Therefore the flux
   // in each bin is proportional to the bin width.
   RealArray flatFlux(.0, nBins);
   Real lowerE = ebins[0];
   for (size_t i=0; i<nBins; ++i)
   {
      Real higherE = ebins[i+1];
      flatFlux[i] = higherE - lowerE;
      lowerE = higherE;
   }
   convolve(flatFlux, dummyError, effArea, dummyError);

   const RealArray& chanEmin = eboundsMin();
   const RealArray& chanEmax = eboundsMax();
   const size_t nChans = chanEmin.size();
   RealArray chanWidths(.0, nChans);
   if (nChans != effArea.size())
   {
      throw RedAlert("Channel size mismatch in MultiBinningResponse::calcEffAreaPerChan");
   }
   chanWidths = chanEmax - chanEmin;
   effArea /= chanWidths;
}

void MultiBinningResponse::setZeroGainEffectiveArea (const RealArray& newEffArea)
{
   for (size_t i=0; i<m_responses.size(); i++) {
     m_responses[i].setZeroGainEffectiveArea(newEffArea);
   }

   if (m_savedEffectiveArea.size())
   {
      if (m_savedEffectiveArea.size() != newEffArea.size())
      {
         throw RedAlert("Size mismatch between old and new effective area array.");
      }
      m_savedEffectiveArea = newEffArea;
   }
}

const string MultiBinningResponse::rmfName () const
{
  // all responses are in the same file but individual sub-responses. if there
  // is only one response then return the stored rmfName otherwise return the name
  // with the {n} suffix removed.
  if ( m_responses.size() == 1 ) {
    return m_responses[0].rmfName();
  } else {
    string name = m_responses[0].rmfName();
    return name.substr(0,name.find_first_of('{'));
  }
}

void MultiBinningResponse::rmfName (const string& value)
{
  // if there is only one response then just set that individual response name to 
  // the input. If there are multiple responses then add the suffix for each 
  // individual response name. Note that it should not be possible for value to
  // have a suffix and there to be multiple responses. 
  if ( m_responses.size() == 1 ) {
    m_responses[0].rmfName(value);
  } else {
    for (size_t i=0; i<m_responses.size(); i++) {
      std::stringstream name;
      name << value << '{' << i+1 << '}';
      m_responses[i].rmfName(name.str());
    }
  }
}

// Additional Declarations
