
// Weight
#include <XSStat/Weight.h>
// TestStatMethod
#include <XSStat/TestStatMethod.h>

#include <XSModel/Model/Model.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/DataUtility.h>
#include <XSModel/Data/Detector/Response.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/Parameter/ModParam.h>
#include <XSModel/Parameter/ResponseParam.h>
#include <XSStat/TestChiSquare.h>
#include <XSStat/PearsonChiSquare.h>
#include <XSStat/EDF.h>
#include <XSStat/EDFVariants.h>
#include <XSStat/Runs.h>
#include <XSstreams.h>
#include <XSContainer.h>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <set>


// Class TestStatMethod::InvalidSpectrumNumber 

TestStatMethod::InvalidSpectrumNumber::InvalidSpectrumNumber (size_t num)
  : RedAlert(" initializing spectra for fit:\n ")  
{
  std::ostringstream message;
  message << " Statistic tried to access spectrum number " << num << " not loaded ";
  reportAndExit(message.str(),-5);
}


// Class TestStatMethod::InvalidParameterError 

TestStatMethod::InvalidParameterError::InvalidParameterError ()
  : RedAlert(" retrieving model derivative array:\n")              
{
  std::ostringstream message;
  message << " Statistic calculation attempted to retrieve parameter derivative not allocated ";
  reportAndExit(message.str(),-5);
}


// Class TestStatMethod::FitNotDefined 

TestStatMethod::FitNotDefined::FitNotDefined (const string& diag)
  : YellowAlert (" no model defined or no data read ")
{
  tcerr << diag << '\n';
}


// Class TestStatMethod::NoSuchSpectrum 

TestStatMethod::NoSuchSpectrum::NoSuchSpectrum (const string& diag)
  : YellowAlert(" spectrum not loaded ")
{
}


// Class TestStatMethod 


TestStatMethod::TestStatMethod (const std::string& name)
  : StatMethod(name),
    m_spectraForStat(),
    m_canCalcNullHyp(false)
{
}


TestStatMethod::~TestStatMethod()
{
}

void TestStatMethod::perform()
{
  doPerform();
}

void TestStatMethod::doPerform()
{
}

void TestStatMethod::initialize ()
{

  doInitialize();
}

void TestStatMethod::doInitialize ()
{

  // clear all and start again.
  deallocate();
  // allocate memory for derivative arrays. If derivatives are not
  // required we don't need to do anything.

  // create proto workspace

  for (size_t i=0; i<m_spectraForStat.size(); ++i)
  {
     const SpectralData* spec = m_spectraForStat[i];
     size_t N = spec->indirectNotice().size();
     m_protoWorkspace.insert(ArrayContainer::value_type(spec->spectrumNumber(),RealArray(0.,N)));  
  }

}

void TestStatMethod::deallocate ()
{

        doDeallocate();
}

void TestStatMethod::doDeallocate ()
{
        m_protoWorkspace.clear(); 
}

void TestStatMethod::reportWarnings ()
{
  doReportWarnings();
}

size_t TestStatMethod::nBinsUsed () const
{
    size_t nBins = 0;
    const size_t nSpec = m_spectraForStat.size();
    for (size_t i=0; i<nSpec; ++i)
    {
       nBins += m_spectraForStat[i]->indirectNotice().size();
    }      
    return nBins;
}

Real TestStatMethod::calcNullHyp(const int nDOF) const
{
  return doCalcNullHyp(nDOF);
}

Real TestStatMethod::doCalcNullHyp(const int nDOF) const
{
  // this should never actually be invoked because any statistic which can
  // calculate the null hypothesis will have its own version of this routine
  return -999.0;
}

TestStatMethod* TestStatMethod::createTestStatMethodObject(const string& name, const int& number, const string& modifier) 
{
  // this method creates and returns the pointer to the appropriate new 
  // object

  static string chi("chi");
  static string pchi("pchi");
  static string ks("ks");
  static string cvm("cvm");
  static string ad("ad");
  static string cusum("cusum");
  static string runs("runs");

  TestStatMethod* stat = 0;
  if ( name == chi.substr(0,name.length()) ) { // chi
    // if a modifier string is included check that it is a valid weighting name
    if ( modifier.length() > 0 ) {
      string weightMethod = modifier;
      if ( Weight::isValidWeightName(weightMethod) == false ) {
	string errStr = modifier + " is not a valid weighting for chi-squared\n";
	throw YellowAlert(errStr);
      }
    }
    stat = new TestChiSquare;
  } else if ( name == pchi.substr(0,name.length()) ) { // pchi
    stat = new PearsonChiSquare;
  } else if ( name == ks.substr(0,name.length()) ) { // ks
    stat = new EDF<K_S>;
  } else if ( name == cvm.substr(0,name.length()) ) { // cvm
    stat = new EDF<C_vM>;
  } else if ( name == ad.substr(0,name.length()) ) {  // ad
    stat = new EDF<A_D>;
  } else if ( name == cusum.substr(0,name.length()) ) { // cusum
    stat = new EDF<CUSUM>;
  } else if ( name == runs.substr(0,name.length()) ) { // runs
    stat = new Runs;
  } else {
    return 0;
  }

  // now the new objects have been created set the number or modifier as appropriate
  if ( number != -999 ) stat->number(number);
  if ( modifier != "" ) stat->modifier(modifier);

  return stat;

}

bool TestStatMethod::isValidStatName(const string& name, const int& numeral, const string& modifier)
{
  bool valid(false);

  // try to create a TestStatMethod object
  try {
    TestStatMethod* stat = createTestStatMethodObject(name, numeral, modifier);
    // if the created stat exists then input is valid
    if (stat) valid = true;
    // tidy up
    delete stat;
  } catch(...) {
    valid = false;
  }

  return valid;
}

string TestStatMethod::statNames ()
{
  return "chi | pchi | ks | cvm | ad | cusum | runs";
}




// Additional Declarations
