//
//  XSPEC12  November 2003
//
//

#include <XSFit/Fit/Fit.h>
#include <XSFit/Fit/StatManager.h>
#include <XSFit/Fit/FitMethod.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSStat/StatMethod.h>
#include <XSStat/FitStatMethod.h>
#include <XSStat/TestStatMethod.h>
#include <XSUser/Global/XSGlobal.h>  
#include <XSUser/Global/Global.h>
#include <XSUser/Handler/HandlerUtils.h>
#include <XSUser/Handler/XSinterface.h>
#include <XSUtil/Parse/XSparse.h>
#include <XSstreams.h>
#include <XSContainer.h>

int
XSGlobal::xsStatistic(ClientData cdata,Tcl_Interp* tclInterp,int objc, Tcl_Obj* CONST  objv[])
{
   StringArray rawArgs;
   HandlerUtils::tclArgsToCpp(objc, objv, rawArgs);
   int status = doStatistic(rawArgs);
   if (!status)
      return globalData->autoSave(TCL_OK);
   else
      return globalData->autoSave(TCL_ERROR);
}

int XSGlobal::doStatistic(const StringArray& rawArgs)
{

   const size_t nArgs = rawArgs.size();
   const char* cmd = "statistic";
   const string statFullName = XSContainer::fit->statManager()->defaultStatMethod()->fullName();
   const string testStatFullName = XSContainer::fit->statManager()->getTestStatMethod()->fullName();
   static RangePair prevRange(0,0);

   // useful here to split the arguments up into range arguments and other strings
   // also check for a "?" as one of the strings
   vector<string> rangeStrings, otherStrings;
   bool isQuery(false);
   const string rangeChars = "0123456789-*,";
   for (size_t iArgs=1; iArgs<nArgs; iArgs++) {
     const string instring = rawArgs[iArgs];
     if ( instring.find_first_not_of(rangeChars) == string::npos ) {
       rangeStrings.push_back(instring);
     } else {
       otherStrings.push_back(instring);
       if ( instring == "?" ) isQuery = true;
     }
   }

   // check for the test statistic being specified. note that this assumes
   // none of the statistic names start with t. The test specifier should
   // be the first entry in otherStrings.

   bool isTest(false);
   if ( otherStrings.size() > 0 ) {
     if ( otherStrings[0].substr(0,1) == "t" || 
	  otherStrings[0].substr(0,1) == "T" ) { 
       otherStrings.erase(otherStrings.begin());
       isTest = true;
     }
   }

   // write information if prompted by the user or not enough arguments given

   if (nArgs == 1 || isQuery || otherStrings.size() == 0) { 
       printDocs(cmd,"?");
       tcout <<"(Current default fit statistic: " 
             << statFullName <<")" <<std::endl;
       tcout <<"(Current test statistic: " 
             << testStatFullName <<")" <<std::endl;
       tcout << std::flush;
       return 0;
   } 

   // Split any modifiers off the statistic name. First check for a numeral
   // and then after that a string

   string statName, modstring("");
   int numeral(-999);
   StatMethod::interpretStatCommandName(otherStrings[0], statName, numeral, modstring);
   otherStrings.erase(otherStrings.begin());

   // check for a modifier as a separate argument following the statistic name
   if ( otherStrings.size() > 0 ) {
     if ( modstring != "" ) {
       tcout << "Incompatible statistic modifiers: both " << modstring << " and " 
	     << otherStrings[0] << " given.";
       return -1;
     }
     modstring = otherStrings[0];
     if ( otherStrings.size() > 1 ) {
       tcout << "Too many arguments: ";
       for (size_t i=1; i<otherStrings.size(); i++ ) tcout << otherStrings[i] << " ";
       tcout << "ignored." << std::endl;
     }
   }
   
   // it is useful to have a pointer to the StatManager singleton
   StatManager* manager = XSContainer::fit->statManager();

   if ( isTest ) {
     // case of a test statistic - there is only one allowed so a range of spectra
     // cannot be specified
     if ( rangeStrings.size() > 0 ) {
       tcout << "***Only one test statistic is allowed, a spectrum number or range cannot be specified." << std::endl;
       return -1;
     }
     // check that the input is valid
     if (!TestStatMethod::isValidStatName(statName, numeral, modstring)) {
       tcout << "***" << statName;
       if ( numeral != -999 ) tcout << " " << numeral;
       if ( modstring.length() > 0 ) tcout << " " << modstring;
       tcout << " is an unrecognized test statistic option." << std::endl;
       printDocs(cmd,"?");
       return -1;
     }
     try {
       // set test statistic method - this will throw if statName is not a valid
       manager->setTestStatMethod(statName, numeral, modstring);
       tcout << "Test statistic is set to: " << manager->getTestStatMethod()->fullName()
	     << std::endl;
     } catch ( YellowAlert& ) {
       printDocs(cmd,"?");
       return -1;   
     }

   } else {
     // case of a fit statistic. first check that the input is valid
     if (!FitStatMethod::isValidStatName(statName, numeral, modstring)) {
       tcout << "***" << statName;
       if ( numeral != -999 ) tcout << " " << numeral;
       if ( modstring.length() > 0 ) tcout << " " << modstring;
       tcout << " is an unrecognized fit statistic option." << std::endl;
       printDocs(cmd,"?");
       return -1;
     }
     // check for an argument giving a range of spectra which will use the statistic
     bool isSettingDefault = false;
     const size_t nSpectra = XSContainer::datasets->numberOfSpectra();
     RangePair limits(1,nSpectra);
     IntegerVector selectedSpecs;
     // Note that nSpectra may be 0
     if ( rangeStrings.size() == 0 ) {
       isSettingDefault = true;
       selectedSpecs.resize(nSpectra);
       for (size_t i=0; i<nSpectra; i++) selectedSpecs[i] = i+1;
     } else {
       // When a range is supplied, don't make the StatMethod the new default.
       if (!nSpectra) {
	 tcout << "***No spectra currently loaded - cannot change their statistics."<<std::endl;
	 return -1;
       }
       if (!prevRange.first) prevRange.first = 1;
       if (!prevRange.second) prevRange.second = nSpectra;
       // This may throw
       try {
	 selectedSpecs = XSparse::getRanges(rangeStrings, prevRange, limits);
       } catch(...) {
	 tcout << "***Invalid spectrum range requested." << std::endl;
	 return -1;
       }
     }

     string statMethodFullName;
     try {
       if ( isSettingDefault ) {
	 manager->setDefaultStatMethod(statName, numeral, modstring);
       }
       for (size_t i=0; i<selectedSpecs.size(); ++i) {
         FitStatMethod* tmpStat = manager->setStatMethodForSpectrum(selectedSpecs[i],statName, numeral, modstring, false);
         if (tmpStat && statMethodFullName.empty())
            statMethodFullName = tmpStat->fullName();
         delete tmpStat;
       }
     } catch ( YellowAlert& ) {
       printDocs(cmd,"?");
       return -1;   
     }

     // Tell the user what is happening.
     if ( isSettingDefault ) {
       tcout << "Default fit statistic is set to: " <<manager->defaultStatMethod()->fullName() 
	     << "\n   This will apply to all current and newly loaded spectra."<<std::endl;
     } else {
       if ( nSpectra > 0 ) {
	 tcout << "The " << statMethodFullName
	       <<" fit statistic will be applied to the specified spectra.\n" 
	       << "   The default fit statistic is: " 
	       << manager->defaultStatMethod()->fullName()<<std::endl;
       }
     }
     

   }

   XSContainer::fit->Update();
   tcout << std::flush;
   return 0;
}





